/*
 * This implements the thrpool event framework interface
 * to the libtevent event framework
 */
#include <private/pprio.h>
#include <talloc.h>
#include <tevent.h>
typedef struct tevent_context ns_event_fw_ctx_t;
typedef struct tevent_fd ns_event_fw_fd_t;
typedef struct tevent_timer ns_event_fw_time_t;
typedef struct tevent_signal ns_event_fw_sig_t;
#define NO_EVENT_TYPEDEFS
#include "ns_event_fw.h"

static void ns_event_fw_add_io(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_mod_io(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_add_timer(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_mod_timer(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_add_signal(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_mod_signal(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_io_event_remove(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_io_event_done(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_timer_event_remove(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_timer_event_done(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_signal_event_remove(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);
static void ns_event_fw_signal_event_done(ns_event_fw_ctx_t *ns_event_fw_ctx, ns_job_t *job);

static ns_job_type_t
event_flags_to_type(uint16_t flags)
{
    ns_job_type_t job_type = 0;
    if (flags & TEVENT_FD_READ) {
        job_type |= NS_JOB_READ;
    }
    if (flags & TEVENT_FD_WRITE) {
        job_type |= NS_JOB_WRITE;
    }
    return job_type;
}

/* this is called by the event main loop when an fd event
   is triggered - this "maps" the event library interface
   to our nspr/thrpool interface */
static void
fd_event_cb(struct tevent_context *ev, struct tevent_fd *fde,
            uint16_t flags, void *arg)
{
    ns_job_t *job = (ns_job_t *)arg;

    PR_ASSERT(arg);

    /* I/O jobs can be timed - if we get an fd event, cancel the timer */
    if (job->ns_event_fw_time) {
        ns_event_fw_timer_event_done(job->ns_event_fw_ctx, job);
    }
    job->output_job_type = event_flags_to_type(flags);
    job->event_cb(job);
}

/* this is called by the event main loop when a timer event
   is triggered - this "maps" the event library interface
   to our nspr/thrpool interface */
static void
timer_event_cb(struct tevent_context *ev, struct tevent_timer *te,
               struct timeval current_time, void *arg)
{
    ns_job_t *job = (ns_job_t *)arg;

    PR_ASSERT(arg);

    /* I/O jobs can be timed - if we get an timer event, cancel the fd */
    if (job->ns_event_fw_fd) {
        ns_event_fw_io_event_done(job->ns_event_fw_ctx, job);
    }
    job->output_job_type = NS_JOB_TIMER;
    job->event_cb(job);
}

/* this is called by the event main loop when a signal event
   is triggered - this "maps" the event library interface
   to our nspr/thrpool interface */
static void
signal_event_cb(struct tevent_context *ev, struct tevent_signal *se,
                int signum, int count, void *siginfo, void *arg)
{
    ns_job_t *job = (ns_job_t *)arg;

    PR_ASSERT(arg);

    job->output_job_type = NS_JOB_SIGNAL;
    job->signal = signum;
    job->event_cb(job);
}

/* convert from job job_type to event fw type */
static uint16_t
job_type_to_flags(ns_job_type_t job_type)
{
    uint16_t flags = 0;
    if (NS_JOB_IS_ACCEPT(job_type) || NS_JOB_IS_READ(job_type)) {
        flags |= TEVENT_FD_READ;
    }
    if (NS_JOB_IS_WRITE(job_type) || NS_JOB_IS_CONNECT(job_type)) {
        flags |= TEVENT_FD_WRITE;
    }
    return flags;
}

static ns_event_fw_ctx_t *
ns_event_fw_init(void)
{
    return tevent_context_init(NULL);
}

static void
ns_event_fw_destroy(ns_event_fw_ctx_t *ns_event_fw_ctx_t)
{
    talloc_free(ns_event_fw_ctx_t);
}

/*
 * remove just removes the event from active consideration
 * by the event framework - the event object may be
 * reactivated by calling mod - event done will actually
 * remove and free the object
 */
static void
ns_event_fw_io_event_remove(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    tevent_fd_set_flags(job->ns_event_fw_fd, 0);
}

static void
ns_event_fw_io_event_done(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    ns_event_fw_io_event_remove(ns_event_fw_ctx, job);
    talloc_free(job->ns_event_fw_fd);
    job->ns_event_fw_fd = NULL;
}

static void
ns_event_fw_timer_event_remove(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    PR_ASSERT(0); /* not supported */
}

static void
ns_event_fw_timer_event_done(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    talloc_free(job->ns_event_fw_time);
    job->ns_event_fw_time = NULL;
}

static void
ns_event_fw_signal_event_remove(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    PR_ASSERT(0); /* not supported */
}

static void
ns_event_fw_signal_event_done(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    talloc_free(job->ns_event_fw_sig);
    job->ns_event_fw_sig = NULL;
}

static void
ns_event_fw_add_io(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    /* set the fields in the event */
    uint16_t flags = job_type_to_flags(job->job_type);

    job->ns_event_fw_fd = tevent_add_fd(ns_event_fw_ctx, ns_event_fw_ctx,
                                        PR_FileDesc2NativeHandle(job->fd),
                                        flags, fd_event_cb, job);
    /* add/schedule the timer event */
    if (job->tv.tv_sec || job->tv.tv_usec) {
        job->ns_event_fw_time = tevent_add_timer(ns_event_fw_ctx, ns_event_fw_ctx,
                                                 job->tv, timer_event_cb, job);
    }
}

static void
ns_event_fw_mod_io(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    uint16_t events = job_type_to_flags(job->job_type);

    if (job->tv.tv_sec || job->tv.tv_usec) {
        ns_event_fw_mod_timer(ns_event_fw_ctx, job);
    }
    tevent_fd_set_flags(job->ns_event_fw_fd, events);
}

static void
ns_event_fw_add_timer(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    job->ns_event_fw_time = tevent_add_timer(ns_event_fw_ctx, ns_event_fw_ctx,
                                             job->tv, timer_event_cb, job);
}

static void
ns_event_fw_mod_timer(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    ns_event_fw_timer_event_done(ns_event_fw_ctx, job);
    ns_event_fw_add_timer(ns_event_fw_ctx, job);
}

static void
ns_event_fw_add_signal(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    job->ns_event_fw_sig = tevent_add_signal(ns_event_fw_ctx, ns_event_fw_ctx,
                                             job->signal, 0, signal_event_cb, job);
}

static void
ns_event_fw_mod_signal(
    ns_event_fw_ctx_t *ns_event_fw_ctx,
    ns_job_t *job
)
{
    ns_event_fw_signal_event_done(ns_event_fw_ctx, job);
    ns_event_fw_add_signal(ns_event_fw_ctx, job);
}

/* returns
   1 - no events to process
   0 - normal termination
   -1 - error
*/
static int
ns_event_fw_loop(ns_event_fw_ctx_t *ns_event_fw_ctx)
{
    int rc = tevent_loop_once(ns_event_fw_ctx);
    if (rc == 0) {
        rc = 1;
    } else {
        rc = -1;
    }
    /* tevent apparently has no normal termination return value */
    return rc;
}

static ns_event_fw_t ns_event_fw_tevent = {
    ns_event_fw_init,
    ns_event_fw_destroy,
    ns_event_fw_loop,
    ns_event_fw_add_io,
    ns_event_fw_mod_io,
    ns_event_fw_add_timer,
    ns_event_fw_mod_timer,
    ns_event_fw_add_signal,
    ns_event_fw_mod_signal,
    ns_event_fw_io_event_done,
    ns_event_fw_timer_event_done,
    ns_event_fw_signal_event_done
};

ns_event_fw_t *
get_event_framework_tevent()
{
    return &ns_event_fw_tevent;
}
