
#include "nspr.h"
#include "pratom.h"
#include "private/pprio.h"
#include "liblfds.h"
#include "ns_event_fw.h"

/*
 * Threadpool
 * PR_Atomic*() operates on PRInt32
 */
struct ns_thrpool_t {
    PRInt32 init_threads;
    PRInt32 max_threads;
    PRInt32 idle_threads;
    PRUint32 stacksize;
    struct stack_state *thread_stack; /* stack of ns_thread_t */
    PRLock *thread_stack_lock; /* to support ns_thread_acquire/release */
    PRCondVar *thread_stack_cv; /* to support ns_thread_acquire/release */
    PRInt32 current_threads;
    PRThread *event_thread; /* the event loop thread */
    struct queue_state *work_q; /* queue of jobs to be executed by the thread pool */
    PRInt32 work_q_size; /* number of elements in work_q */
    PRLock *work_q_lock;
    PRCondVar *work_q_cv;
    struct queue_state *event_q; /* queue of jobs to return to the event loop */
    PRInt32 event_q_size; /* number of elements in event_q */
    PRFileDesc *event_q_wakeup_pipe_read;
    PRFileDesc *event_q_wakeup_pipe_write;
    ns_job_t *event_q_wakeup_job;
    PRInt32 shutdown;
    PRInt32 shutdown_event_loop;
    ns_event_fw_t *ns_event_fw;
    ns_event_fw_ctx_t *ns_event_fw_ctx;
    /* stats */
    PRInt32 min_idle_threads; /* minimum number of idle threads */
    PRInt32 max_work_q_size; /* maximum size of work q */
    PRInt32 max_event_q_size; /* maximum size of event q */
};

struct ns_thread_t {
    PRThread *thr; /* the thread */
    struct ns_thrpool_t *tp; /* pointer back to thread pool */
    struct queue_state *work_q; /* queue of jobs to be executed by this thread only */
};

#define ATOMIC_VAL_IS_ZERO(val) 0 == PR_AtomicAdd((val), 0)

static void
thread_stack_wait(ns_thrpool_t *tp)
{
    PR_Lock(tp->thread_stack_lock);
    PR_WaitCondVar(tp->thread_stack_cv, PR_INTERVAL_NO_TIMEOUT);
    PR_Unlock(tp->thread_stack_lock);
}

static void
thread_stack_notify(ns_thread_t *thr)
{
    PR_Lock(thr->tp->thread_stack_lock);
    PR_NotifyCondVar(thr->tp->thread_stack_cv);
    PR_Unlock(thr->tp->thread_stack_lock);
}

static void
work_q_wait(ns_thrpool_t *tp)
{
    PR_Lock(tp->work_q_lock);
    PR_WaitCondVar(tp->work_q_cv, PR_INTERVAL_NO_TIMEOUT);
    PR_Unlock(tp->work_q_lock);
}

static void
work_q_notify(ns_job_t *job)
{
    if (NS_JOB_IS_THREAD(job->job_type)) {
        queue_enqueue(job->tp->work_q, job);
        (void)PR_AtomicIncrement(&job->tp->work_q_size);
        if (job->tp->work_q_size > job->tp->max_work_q_size) {
            job->tp->max_work_q_size = job->tp->work_q_size;
        }
        PR_Lock(job->tp->work_q_lock);
        PR_NotifyCondVar(job->tp->work_q_cv);
        PR_Unlock(job->tp->work_q_lock);
        PR_Sleep(PR_INTERVAL_NO_WAIT); /* yield to allow worker thread to pick up job */
    } else {
        /* execute in same thread */
        ns_job_func_t jobfunc = job->func;
        jobfunc(job);
    }
}

/*
 * worker thread function
 */
static void
worker_thread_func(void *arg)
{
    ns_thread_t *thr = (ns_thread_t *)arg;
    ns_thrpool_t *tp = thr->tp;

    /*
     * Execute jobs until shutdown is set and the queues are empty.
     */
    while (ATOMIC_VAL_IS_ZERO(&tp->shutdown) || !(ATOMIC_VAL_IS_ZERO(&tp->work_q_size)) ||
           !(ATOMIC_VAL_IS_ZERO(&tp->event_q_size))) {
        ns_job_t *job = NULL;
        /* NGK - could this cause a tight loop if shutdown is set and nothing is in the work queue,
         * but we have a queued event?  I don't think this scenario would happen very long, so
         * maybe this is OK? */
        while (!queue_dequeue(thr->work_q, (void **)&job) &&
               !queue_dequeue(tp->work_q, (void **)&job) &&
               ATOMIC_VAL_IS_ZERO(&tp->shutdown)) {
            work_q_wait(tp);
        }
        if (job) {
            ns_job_func_t jobfunc = job->func;
            (void)PR_AtomicDecrement(&tp->work_q_size);
            (void)PR_AtomicDecrement(&tp->idle_threads);
            if (tp->idle_threads < tp->min_idle_threads) {
                tp->min_idle_threads = tp->idle_threads;
            }
            jobfunc(job);
            (void)PR_AtomicIncrement(&tp->idle_threads);
        }
    }
    (void)PR_AtomicDecrement(&tp->current_threads);
    (void)PR_AtomicDecrement(&tp->idle_threads);
}

static void
internal_ns_job_done(ns_job_t *job)
{
    (void)PR_AtomicIncrement(&job->delete_state);
    if (job->ns_event_fw_fd) {
        job->tp->ns_event_fw->ns_event_fw_io_event_done(job->tp->ns_event_fw_ctx, job);
    }
    if (job->ns_event_fw_time) {
        job->tp->ns_event_fw->ns_event_fw_timer_event_done(job->tp->ns_event_fw_ctx, job);
    }
    if (job->ns_event_fw_sig) {
        job->tp->ns_event_fw->ns_event_fw_signal_event_done(job->tp->ns_event_fw_ctx, job);
    }

    if (job->fd && !NS_JOB_IS_PRESERVE_FD(job->job_type)) {
        PR_Close(job->fd);
    } /* else application is responsible for fd */
    PR_DELETE(job);
}

/*
 * Add a new event, or modify or delete an existing event
 */
static void
update_event(ns_job_t *job)
{
    if (job->delete_state == NS_JOB_NEEDS_DELETE) {
        internal_ns_job_done(job);
    } else if (NS_JOB_IS_IO(job->job_type) || job->ns_event_fw_fd) {
        if (!job->ns_event_fw_fd) {
            job->tp->ns_event_fw->ns_event_fw_add_io(job->tp->ns_event_fw_ctx, job);
        } else {
            job->tp->ns_event_fw->ns_event_fw_mod_io(job->tp->ns_event_fw_ctx, job);
        }
    } else if (NS_JOB_IS_TIMER(job->job_type) || job->ns_event_fw_time) {
        if (!job->ns_event_fw_time) {
            job->tp->ns_event_fw->ns_event_fw_add_timer(job->tp->ns_event_fw_ctx, job);
        } else {
            job->tp->ns_event_fw->ns_event_fw_mod_timer(job->tp->ns_event_fw_ctx, job);
        }
    } else if (NS_JOB_IS_SIGNAL(job->job_type) || job->ns_event_fw_sig) {
        if (!job->ns_event_fw_sig) {
            job->tp->ns_event_fw->ns_event_fw_add_signal(job->tp->ns_event_fw_ctx, job);
        } else {
            job->tp->ns_event_fw->ns_event_fw_mod_signal(job->tp->ns_event_fw_ctx, job);
        }
    }

    return;
}

static void
event_q_wait(ns_thrpool_t *tp)
{
    /* unused for now */
    /* the main event loop will do our "waiting" - waiting for events
       to happen, or we can trigger an event to "wakeup" the event
       loop (see event_q_notify) */
}

static void
event_q_wake(ns_thrpool_t *tp)
{
    PRInt32 len;

    len = PR_Write(tp->event_q_wakeup_pipe_write, "a", 1);
    if (1 != len) {
        PRErrorCode prerr = PR_GetError();
        fprintf(stderr, "Error: could not write wakeup pipe: %d:%s\n",
                prerr, PR_ErrorToString(prerr, PR_LANGUAGE_I_DEFAULT));
    }
    PR_Sleep(PR_INTERVAL_NO_WAIT); /* yield to allow event thread to pick up event */
}

static void
event_q_notify(ns_job_t *job, int cur_is_thread)
{
    /* if we are being called from a thread other than the
       event loop thread, we have to notify that thread to
       perform the event work */
    if (cur_is_thread) {
        /* The event loop may be waiting for events, and may wait a long
           time by default if there are no events to process - since we
           want to add an event, we have to "wake up" the event loop by
           posting an event - this will cause the wakeup_cb to be called
           which will empty the event_q and add all of the events
        */
        ns_thrpool_t *tp = job->tp;
        /* NOTE: once job is queued, it may be deleted immediately in
         * another thread, if the event loop picks up the deletion
         * job before we can notify it below - so make sure not to
         * refer to job after the enqueue.
         */
        queue_enqueue(tp->event_q, job);
        (void)PR_AtomicIncrement(&tp->event_q_size);
        if (tp->event_q_size > tp->max_event_q_size) {
            tp->max_event_q_size = tp->event_q_size;
        }
        event_q_wake(tp);
    } else {
        /* otherwise, if we are being run from the same thread as the event
           loop thread, we can just update the event here */
        update_event(job);
    }
}

/* This is run inside the event loop thread, and only in the
   event loop thread
   This function pulls the io/timer/signal event requests off
   the request queue, formats the events in the format required
   by the event framework, and adds them
*/
static void
get_new_event_requests(ns_thrpool_t *tp)
{
    ns_job_t *job = NULL;
    while (queue_dequeue(tp->event_q, (void **)&job)) {
        (void)PR_AtomicDecrement(&tp->event_q_size);
        update_event(job);
    }
}

static void
event_loop_thread_func(void *arg)
{
    ns_thrpool_t *tp = (ns_thrpool_t *)arg;
    while (ATOMIC_VAL_IS_ZERO(&tp->shutdown_event_loop)) {
        int rc;
        /* get new event requests */
        get_new_event_requests(tp);
        /* process events */
        /* return 1 - no events ; 0 - normal exit ; -1 - error */
        rc = tp->ns_event_fw->ns_event_fw_loop(tp->ns_event_fw_ctx);
        if (rc == -1) { /* error */
        } else if (rc == 0) { /* exiting */
        } else { /* no events to process */
            event_q_wait(tp);
        }
    }
}

/*
 * The event framework calls this function when it receives an
 * event - the event framework does all of the work to map
 * the framework flags, etc. into the job
 * This function is called only in the event loop thread
 * if the THREAD flag is set, this means the job is to be
 * handed off to the work q - otherwise, execute it now
 * in this thread (the event loop thread)
 * the function must be careful not to block the event
 * loop thread or starvation will occur
 */
static void
event_cb(ns_job_t *job)
{
    if (NS_JOB_IS_THREAD(job->job_type)) {
        work_q_notify(job);
    } else {
        job->func(job);
    }
}

static void
wakeup_cb(ns_job_t *job)
{
    PRInt32 len;
    char buf[1];

    len = PR_Read(job->tp->event_q_wakeup_pipe_read, buf, 1);
    if (1 != len) {
        PRErrorCode prerr = PR_GetError();
        fprintf(stderr, "Error: could not read wakeup pipe: %d:%s\n",
                prerr, PR_ErrorToString(prerr, PR_LANGUAGE_I_DEFAULT));
    }
    /* wakeup_cb is usually called because a worker thread has posted a new
       event we need to add - get all new event requests */
    get_new_event_requests(job->tp);
}

/* convenience function for the event fw to use to allocate
   space for its event fw event object
   it is assumed that eventually job will have a memory region/arena
   to use
   these functions will be called in the event loop thread from
   the event framework function that adds a new event
*/
static void *
alloc_event_context(size_t size, ns_job_t *job)
{
    return malloc(size);
}

static void
free_event_context(void *ev_ctx, ns_job_t *job)
{
    free(ev_ctx);
}

static ns_job_t *
new_ns_job(ns_thrpool_t *tp, PRFileDesc *fd, ns_job_type_t job_type, ns_job_func_t func, struct ns_job_data_t *data)
{
    ns_job_t *job = PR_NEWZAP(ns_job_t);
    job->tp = tp;
    job->ns_event_fw_ctx = tp->ns_event_fw_ctx;
    job->fd = fd;
    job->func = func;
    job->data = data;
    job->alloc_event_context = alloc_event_context;
    job->free_event_context = free_event_context;
    job->event_cb = event_cb;
    job->job_type = job_type;
    return job;
}

static ns_job_t *
alloc_io_context(ns_thrpool_t *tp, PRFileDesc *fd, ns_job_type_t job_type,
        ns_job_func_t func, struct ns_job_data_t *data)
{
    ns_job_t *job = new_ns_job(tp, fd, job_type, func, data);

    return job;
}

static ns_job_t *
alloc_timeout_context(ns_thrpool_t *tp, struct timeval *tv, ns_job_type_t job_type,
        ns_job_func_t func, struct ns_job_data_t *data)
{
    ns_job_t *job = new_ns_job(tp, NULL, NS_JOB_TIMER | job_type, func, data);
    job->tv = *tv;

    return job;
}

static ns_job_t *
alloc_signal_context(ns_thrpool_t *tp, PRInt32 signum, ns_job_type_t job_type,
        ns_job_func_t func, struct ns_job_data_t *data)
{
    ns_job_t *job = new_ns_job(tp, NULL, NS_JOB_SIGNAL | job_type, func, data);
    job->signal = signum;

    return job;
}

void
ns_job_done(ns_job_t *job)
{
    PR_ASSERT(!job->delete_state);
    (void)PR_AtomicIncrement(&job->delete_state);
    event_q_notify(job, NS_JOB_IS_THREAD(job->job_type));
}

/* queue a file descriptor to listen for and accept new connections */
PRStatus
ns_add_io_job(ns_thrpool_t *tp, PRFileDesc *fd, ns_job_type_t job_type,
        ns_job_func_t func, void *data, ns_job_t **job)
{
    ns_job_t *_job = NULL;

    /* Don't allow a job to be added if the threadpool is being shut down. */
    if (ns_thrpool_is_shutdown(tp)) {
        return PR_FAILURE;
    }

    /* Don't allow an accept job to be run outside of the event thread. 
     * We do this so a listener job won't shut down while still processing
     * current connections in other threads.
     * TODO: Need to be able to have multiple threads accept() at the same time
     * This is fine - just have to remove the listen job in the polling thread
     * immediately after receiving notification - then call the job to do the
     * accept(), which will add back the persistent listener job immediately after
     * doing the accept()
     * This will be a combination of a non-threaded job and a threaded job
     *
     */
    if (NS_JOB_IS_ACCEPT(job_type) && NS_JOB_IS_THREAD(job_type)) {
        return PR_FAILURE;
    }

    /* get an event context for an accept */
    _job = alloc_io_context(tp, fd, job_type, func, data);

    event_q_notify(_job, NS_JOB_IS_THREAD(job_type));

    /* fill in a pointer to the job for the caller if requested */
    if (job) {
        *job = _job;
    }

    return _job ? PR_SUCCESS : PR_FAILURE;
}

PRStatus
ns_add_timeout_job(ns_thrpool_t *tp, struct timeval *tv, ns_job_type_t job_type,
        ns_job_func_t func, void *data, ns_job_t **job)
{
    ns_job_t *_job = NULL;

    /* Don't allow a job to be added if the threadpool is being shut down. */
    if (ns_thrpool_is_shutdown(tp)) {
        return PR_FAILURE;
    }

    /* get an event context for a timer job */
    _job = alloc_timeout_context(tp, tv, job_type, func, data);

    event_q_notify(_job, NS_JOB_IS_THREAD(_job->job_type));

    /* fill in a pointer to the job for the caller if requested */
    if (job) {
        *job = _job;
    }

    return _job ? PR_SUCCESS : PR_FAILURE;
}

/* queue a file descriptor to listen for and accept new connections */
PRStatus
ns_add_io_timeout_job(ns_thrpool_t *tp, PRFileDesc *fd, struct timeval *tv,
        ns_job_type_t job_type, ns_job_func_t func, void *data, ns_job_t **job)
{
    ns_job_t *_job = NULL;

    /* Don't allow a job to be added if the threadpool is being shut down. */
    if (ns_thrpool_is_shutdown(tp)) {
        return PR_FAILURE;
    }

    /* Don't allow an accept job to be run outside of the event thread.
     * We do this so a listener job won't shut down while still processing
     * current connections in other threads.
     * TODO: Need to be able to have multiple threads accept() at the same time
     * This is fine - just have to remove the listen job in the polling thread
     * immediately after receiving notification - then call the job to do the
     * accept(), which will add back the persistent listener job immediately after
     * doing the accept()
     * This will be a combination of a non-threaded job and a threaded job
     *
     */
    if (NS_JOB_IS_ACCEPT(job_type) && NS_JOB_IS_THREAD(job_type)) {
        return PR_FAILURE;
    }

    /* get an event context for an accept */
    _job = alloc_io_context(tp, fd, job_type|NS_JOB_TIMER, func, data);
    _job->tv = *tv;

    event_q_notify(_job, NS_JOB_IS_THREAD(job_type));

    /* fill in a pointer to the job for the caller if requested */
    if (job) {
        *job = _job;
    }

    return _job ? PR_SUCCESS : PR_FAILURE;
}

PRStatus
ns_add_signal_job(ns_thrpool_t *tp, PRInt32 signum, ns_job_type_t job_type,
        ns_job_func_t func, void *data, ns_job_t **job)
{
    ns_job_t *_job = NULL;

    /* Don't allow a job to be added if the threadpool is being shut down. */
    if (ns_thrpool_is_shutdown(tp)) {
        return PR_FAILURE;
    }

    /* get an event context for a signal job */
    _job = alloc_signal_context(tp, signum, job_type, func, data);

    event_q_notify(_job, NS_JOB_IS_THREAD(_job->job_type));

    /* fill in a pointer to the job for the caller if requested */
    if (job) {
        *job = _job;
    }

    return _job ? PR_SUCCESS : PR_FAILURE;
}

PRStatus
ns_add_job(ns_thrpool_t *tp, struct ns_thread_t *thr, ns_job_type_t job_type, ns_job_func_t func, void *data)
{
    ns_job_t *job;

    /* Don't allow a job to be added if the threadpool is being shut down. */
    if (ns_thrpool_is_shutdown(tp)) {
        return PR_FAILURE;
    }

    job = new_ns_job(tp, NULL, job_type, func, data);
    if (job) {
        work_q_notify(job);
        return PR_SUCCESS;
    }
    return PR_FAILURE;
}

void *
ns_job_get_data(struct ns_job_t *job)
{
    return job->data;
}

ns_thrpool_t *
ns_job_get_tp(struct ns_job_t *job)
{
    return job->tp;
}

ns_job_type_t
ns_job_get_output_type(struct ns_job_t *job)
{
    return job->output_job_type;
}

/*
 * This is a convenience function - use if you need to re-arm the same event
 * usually not needed for persistent jobs
 */
void
ns_job_rearm(ns_job_t *job)
{
    event_q_notify(job, NS_JOB_IS_THREAD(job->job_type));
}

void
ns_job_modify(ns_job_t *job, ns_job_type_t job_type)
{
    ns_job_type_t cur_job_type = job->job_type;
    job->job_type = job_type;
    /* if we are converting a job from an event job to a non-event job, or
       converting from a non-event to an event job, or
       changing the type of event (e.g. from read to write)
       we have to notify the event fw
       we need to let the event q know if the current job is threaded and
       if the new job is threaded
    */
    if (NS_JOB_IS_IO(job_type) || NS_JOB_IS_TIMER(job_type) || NS_JOB_IS_SIGNAL(job_type) ||
        NS_JOB_IS_IO(cur_job_type) || NS_JOB_IS_TIMER(cur_job_type) || NS_JOB_IS_SIGNAL(cur_job_type)) {
        event_q_notify(job, NS_JOB_IS_THREAD(cur_job_type));
    }

    if (!(NS_JOB_IS_IO(job_type) || NS_JOB_IS_TIMER(job_type) || NS_JOB_IS_SIGNAL(job_type))) {
        /* if this is a non event task, just queue it on the work q */
        work_q_notify(job);
    }
}

static void
ns_thrpool_delete(ns_thrpool_t *tp)
{
    PR_DELETE(tp);
}

static ns_thrpool_t *
ns_thrpool_alloc(void)
{
    ns_thrpool_t *tp;

    tp = PR_NEWZAP(ns_thrpool_t);
    if (NULL == tp) {
        goto failed;
    }

    return tp;
failed:
    ns_thrpool_delete(tp);
    PR_SetError(PR_OUT_OF_MEMORY_ERROR, 0);
    return NULL;
}

/* libevent does not make public the file descriptors used to
   wakeup epoll_wait - you have to send a signal - instead of
   that just create our own wakeup pipe */
static void
setup_event_q_wakeup(ns_thrpool_t *tp)
{
    ns_job_t *job;
    PRSocketOptionData prsod = {PR_SockOpt_Nonblocking, {PR_TRUE}};
    PR_CreatePipe(&tp->event_q_wakeup_pipe_read,
                  &tp->event_q_wakeup_pipe_write);
    PR_SetSocketOption(tp->event_q_wakeup_pipe_read, &prsod);
    PR_SetSocketOption(tp->event_q_wakeup_pipe_write, &prsod);
    /* wakeup events are processed inside the event loop thread */
    job = alloc_io_context(tp, tp->event_q_wakeup_pipe_read,
                            NS_JOB_READ|NS_JOB_PERSIST,
                            wakeup_cb, NULL);
    tp->ns_event_fw->ns_event_fw_add_io(tp->ns_event_fw_ctx, job);
    /* Stash the wakeup job in tp so we can release it later. */
    tp->event_q_wakeup_job = job;
}

ns_thrpool_t *
ns_thrpool_new(PRInt32 initial_threads, PRInt32 max_threads, PRUint32 stacksize, PRUint32 queue_size)
{
    ns_thrpool_t *tp = NULL;
    ns_thread_t *thr;
    PRThread *event_thr;
    int ii;

    tp = ns_thrpool_alloc();
    if (NULL == tp) {
        goto failed;
    }

    tp->init_threads = initial_threads;
    /* NGK - how does the thread pool increase the threads if max_threads is set higher than initial_threads? */
    tp->max_threads = max_threads;
    tp->stacksize = stacksize;
    if (!stack_new(&tp->thread_stack, max_threads) || !tp->thread_stack) {
        goto failed;
    }

    if (!queue_new(&tp->work_q, queue_size) || !tp->work_q) {
        goto failed;
    }
    if (!(tp->work_q_lock = PR_NewLock())) {
        goto failed;
    }
    if (!(tp->work_q_cv = PR_NewCondVar(tp->work_q_lock))) {
        goto failed;
    }

    if (!queue_new(&tp->event_q, queue_size) || !tp->event_q) {
        goto failed;
    }

    /* NGK TODO - add tevent vs. libevent switch */
    /* tp->ns_event_fw = get_event_framework_tevent(); */
    tp->ns_event_fw = get_event_framework_event();
    tp->ns_event_fw_ctx = tp->ns_event_fw->ns_event_fw_init();

    setup_event_q_wakeup(tp);

    if (!(tp->thread_stack_lock = PR_NewLock())) {
        goto failed;
    }
    if (!(tp->thread_stack_cv = PR_NewCondVar(tp->thread_stack_lock))) {
        goto failed;
    }
    for (ii = 0; ii < initial_threads; ++ii) {
        thr = PR_NEW(ns_thread_t);
        PR_ASSERT(thr);
        if (!queue_new(&thr->work_q, queue_size) || !thr->work_q) {
            goto failed;
        }
        thr->tp = tp;
        thr->thr = PR_CreateThread(PR_USER_THREAD, worker_thread_func,
                                   thr, PR_PRIORITY_NORMAL, PR_GLOBAL_THREAD,
                                   PR_JOINABLE_THREAD, stacksize);
        PR_ASSERT(thr->thr);
        stack_push(tp->thread_stack, thr);
    }
    tp->current_threads = tp->idle_threads = tp->min_idle_threads = initial_threads;

    event_thr = PR_CreateThread(PR_USER_THREAD, event_loop_thread_func,
                                tp, PR_PRIORITY_NORMAL, PR_GLOBAL_THREAD,
                                PR_JOINABLE_THREAD, stacksize);
    PR_ASSERT(event_thr);

    /* We keep the event thread separate from the stack of worker threads. */
    tp->event_thread = event_thr;

    return tp;
failed:
    ns_thrpool_destroy(tp);
    return NULL;
}

static void
print_stats(struct ns_thrpool_t *tp)
{
    printf("min idle threads: %d\n", tp->min_idle_threads);
    printf("max work q size: %d\n", tp->max_work_q_size);
    printf("max event q size: %d\n", tp->max_event_q_size);
}

void
ns_thrpool_destroy(struct ns_thrpool_t *tp)
{
    if (tp) {
        print_stats(tp);
        /* Set the flag to shutdown the event loop. */
        PR_AtomicSet(&tp->shutdown_event_loop, 1);

        /* Finish the event queue wakeup job.  This has the
         * side effect of waking up the event loop thread, which
         * will cause it to exit since we set the event loop
         * shutdown flag.  Fake the job to be a threaded job
         * so that we can run it from outside the event loop,
         * and use it to wake up the event loop.
         */
        tp->event_q_wakeup_job->job_type |= NS_JOB_THREAD;
        ns_job_done(tp->event_q_wakeup_job);

        /* Wait for the event thread to finish before we free the
         * internals of tp. */
        PR_JoinThread(tp->event_thread);

        /* Free the thread stack. */
        if (tp->thread_stack) {
            /* The stack should be empty if ns_thrpool_wait() returned successfully
             * already.  For this reason, we don't need to supply a data delete callback. */
            stack_delete(tp->thread_stack, NULL, NULL);
            tp->thread_stack = NULL;
        }

        /* Free the work queue. */
        if (tp->work_q) {
            struct ns_job_t *job = NULL;
            int njobs = 0;
            while (queue_dequeue(tp->work_q, (void **)&job) && job) {
                njobs++;
                ns_job_done(job);
            }
            fprintf(stderr, "ns_thrpool_destroy: destroyed [%d] work_q jobs\n", njobs);
            queue_delete(tp->work_q, NULL, NULL);
            tp->work_q = NULL;
        }

        /* Free the work queue lock. */
        if (tp->work_q_lock) {
            PR_DestroyLock(tp->work_q_lock);
            tp->work_q_lock = NULL;
        }

        /* Free the work queue condition variable. */
        if (tp->work_q_cv) {
            PR_DestroyCondVar(tp->work_q_cv);
            tp->work_q_cv = NULL;
        }

        /* Free the thread stack lock. */
        if (tp->thread_stack_lock) {
            PR_DestroyLock(tp->thread_stack_lock);
            tp->thread_stack_lock = NULL;
        }

        /* Free the thread stack condition variable. */
        if (tp->thread_stack_cv) {
            PR_DestroyCondVar(tp->thread_stack_cv);
            tp->thread_stack_cv = NULL;
        }

        /* Free the event queue. */
        if (tp->event_q) {
            struct ns_job_t *job = NULL;
            int njobs = 0;
            while (queue_dequeue(tp->event_q, (void **)&job) && job) {
                njobs++;
                ns_job_done(job);
            }
            fprintf(stderr, "ns_thrpool_destroy: destroyed [%d] event_q jobs\n", njobs);
            queue_delete(tp->event_q, NULL, NULL);
            tp->event_q = NULL;
        }

        /* Free the event queue wakeup pipe. */
        if (tp->event_q_wakeup_pipe_read) {
            PR_Close(tp->event_q_wakeup_pipe_read);
            tp->event_q_wakeup_pipe_read = NULL;
        }

        if (tp->event_q_wakeup_pipe_write) {
            PR_Close(tp->event_q_wakeup_pipe_write);
            tp->event_q_wakeup_pipe_write = NULL;
        }

        /* Free the event framework context. */
        if (tp->ns_event_fw_ctx) {
            tp->ns_event_fw->ns_event_fw_destroy(tp->ns_event_fw_ctx);
            tp->ns_event_fw_ctx = NULL;
        }

        /* Free the thread pool struct itself. */
        ns_thrpool_delete(tp);
    }
}

/* Triggers the pool of worker threads to shutdown after finishing the remaining work. 
 * This must be run in a worker thread, not the event thread.  Running it in the event
 * thread could cause a deadlock. */
void
ns_thrpool_shutdown(struct ns_thrpool_t *tp)
{
    /* Set the shutdown flag.  This will cause the worker
     * threads to exit after they finish all remaining work. */
    PR_AtomicSet(&tp->shutdown, 1);

    /* Wake up the idle worker threads so they can exit. */
    PR_Lock(tp->work_q_lock);
    PR_NotifyAllCondVar(tp->work_q_cv);
    PR_Unlock(tp->work_q_lock);
}

PRInt32
ns_thrpool_is_shutdown(struct ns_thrpool_t *tp)
{
    return ATOMIC_VAL_IS_ZERO(&tp->shutdown) ? 0 : 1;
}

PRStatus
ns_thrpool_wait(ns_thrpool_t *tp)
{
    PRStatus retval = PR_SUCCESS;
    ns_thread_t *thr;
    while (stack_pop(tp->thread_stack, (void **)&thr) && thr) {
        int njobs = 0;
        struct ns_job_t *job = NULL;
        PRStatus rc = PR_JoinThread(thr->thr);
        if (rc != PR_SUCCESS) {
            /* NGK TODO - this is unused right now. */
            PRErrorCode prerr = PR_GetError();
            (void)prerr;
        }
        while (queue_dequeue(thr->work_q, (void **)&job) && job) {
            njobs++;
            ns_job_done(job);
        }
        fprintf(stderr, "ns_thrpool_wait: thread [0x%p] deleted [%d] jobs\n",
                thr, njobs);
        queue_delete(thr->work_q, NULL, NULL);
    }

    return retval;
}

struct ns_thread_t *
ns_thread_acquire(struct ns_thrpool_t *tp, PRBool block)
{
    /* get a thread from the active queue */
    struct ns_thread_t *thr = NULL;

    if (ATOMIC_VAL_IS_ZERO(&tp->shutdown)) {
        while (!stack_pop(tp->thread_stack, (void **)&thr) &&
               ATOMIC_VAL_IS_ZERO(&tp->shutdown)) {
            if (block) {
                thread_stack_wait(tp);
            }
        }
    }
    return thr;
}

void
ns_thread_release(struct ns_thread_t *thr)
{
    stack_push(thr->tp->thread_stack, thr);
    thread_stack_notify(thr);
}
