#ifndef NS_THRPOOL_H
#define NS_THRPOOL_H

#include "nspr.h"

#define PR_WOULD_BLOCK(iii) (iii == PR_PENDING_INTERRUPT_ERROR) || (iii == PR_WOULD_BLOCK_ERROR)

struct ns_thrpool_t;
typedef struct ns_thrpool_t ns_thrpool_t;

struct ns_thread_t;
typedef struct ns_thread_t ns_thread_t;

struct ns_job_t;

typedef void (*ns_job_func_t)(struct ns_job_t *);

/* NGK TODO - need to add doxygen blocks describing how
 * each of these types should be used. */
/* these are the types of jobs we know about */
#define NS_JOB_NONE 0x0
#define NS_JOB_ACCEPT 0x1
#define NS_JOB_CONNECT 0x2
#define NS_JOB_READ 0x4
#define NS_JOB_WRITE 0x8
#define NS_JOB_TIMER 0x10
#define NS_JOB_SIGNAL 0x20
#define NS_JOB_PERSIST 0x40
/* NS_JOB_THREAD allows you to specify if you want a
   job to run threaded or not - if not, the job runs in
   the same thread as the event loop thread - care must be taken
   because this is the only event processing thread - a
   long running job may block all other I/O and timer
   processing - if not threaded, best to use with non-blocking I/O that
   either completes or returns EAGAIN/EWOULDBLOCK
*/
#define NS_JOB_THREAD 0x80
/* If the job->fd is managed elsewhere, this tells
 * ns_job_done not to close the fd, and the application
 * will take care of the fd lifecycle.
 */
#define NS_JOB_PRESERVE_FD 0x100
typedef unsigned short ns_job_type_t;

#define NS_JOB_IS_ACCEPT(eee) ((eee)&NS_JOB_ACCEPT)
#define NS_JOB_IS_READ(eee) ((eee)&NS_JOB_READ)
#define NS_JOB_IS_CONNECT(eee) ((eee)&NS_JOB_CONNECT)
#define NS_JOB_IS_WRITE(eee) ((eee)&NS_JOB_WRITE)
#define NS_JOB_IS_TIMER(eee) ((eee)&NS_JOB_TIMER)
#define NS_JOB_IS_SIGNAL(eee) ((eee)&NS_JOB_SIGNAL)
#define NS_JOB_IS_PERSIST(eee) ((eee)&NS_JOB_PERSIST)
#define NS_JOB_IS_IO(eee) (NS_JOB_IS_ACCEPT(eee)||NS_JOB_IS_READ(eee)||NS_JOB_IS_CONNECT(eee)||NS_JOB_IS_WRITE(eee))
#define NS_JOB_IS_THREAD(eee) ((eee)&NS_JOB_THREAD)
#define NS_JOB_IS_PRESERVE_FD(eee) ((eee)&NS_JOB_PRESERVE_FD)

#define NS_JOB_SET_READ(eee) ((eee) |= NS_JOB_READ)
#define NS_JOB_SET_WRITE(eee) ((eee) |= NS_JOB_WRITE)
#define NS_JOB_SET_THREAD(eee) ((eee) |= NS_JOB_THREAD)
#define NS_JOB_UNSET_READ(eee) ((eee) &= ~NS_JOB_READ)
#define NS_JOB_UNSET_WRITE(eee) ((eee) &= ~NS_JOB_WRITE)
#define NS_JOB_UNSET_THREAD(eee) ((eee) &= ~NS_JOB_THREAD)

/* NGK TODO - document how this is supposed to be used by the job callback functions. */
void ns_job_done(struct ns_job_t *job);

/**
 * Adds an I/O job to the thread pool's work queue
 *
 * \param tp The thread pool whose work queue you want to add an I/O job to.
 * \param fd The file descriptor to use for I/O.
 * \param job_type A set of flags that indicates the job type.
 * \param func The callback function for a worker thread to call when processing the job.
 * \param data Arbitrary data that will be available to the job callback function.
 * \param job The address of a job pointer that will be filled in once the job is allocated.
 *            \c NULL can be passed if a pointer to the job is not needed.
 * \return \c PR_SUCCESS upon success.
 * \return \c PR_FAILURE upon failure.
 * \warn The thread pool will not allow a job to be added when it has been signaled
 *       to shutdown.
 */
PRStatus ns_add_io_job(struct ns_thrpool_t *tp,
                       PRFileDesc *fd,
                       ns_job_type_t job_type,
                       ns_job_func_t func,
                       void *data,
                       struct ns_job_t **job);

/**
 * Adds a timeout job to the thread pool's work queue
 *
 * \param tp The thread pool whose work queue you want to add a timeout job to.
 * \param tv The timer that needs to expire before triggering the callback function.
 * \param job_type A set of flags that indicates the job type.
 * \param func The callback function for a worker thread to call when processing the job.
 * \param data Arbitraty data that will be available to the job callback function.
 * \param job The address of a job pointer that will be filled in once the job is allocated.
 *            \c NULL can be passed if a pointer to the job is not needed.
 * \return \c PR_SUCCESS upon success.
 * \return \c PR_FAILURE upon failure.
 * \warn The thread pool will not allow a job to be added when it has been signaled
 *       to shutdown.
 */
PRStatus ns_add_timeout_job(struct ns_thrpool_t *tp,
                            struct timeval *tv,
                            ns_job_type_t job_type,
                            ns_job_func_t func,
                            void *data,
                            struct ns_job_t **job);

/**
 * Adds an I/O job to the thread pool's work queue with a timeout.
 * The callback func should test the type of event that triggered
 * the callback using ns_job_get_output_type(job) to get the
 * ns_job_type_t, then use NS_JOB_IS_TIMER(output_type) to see if
 * this callback was triggered by a timer event.  This is useful if
 * you want to perform some sort of I/O, but you require that I/O
 * must happen in a certain amount of time, or it is considered an
 * I/O timeout condition.
 * NOTE: This is like adding an I/O job, with an optional timeout.
 * This is not like adding a timeout job with an additional I/O
 * event component.  This depends on the underlying event framework
 * having the ability to have a timed I/O job.  For example, libevent
 * I/O events can have a timeout.
 *
 * \param tp The thread pool whose work queue you want to add an I/O job to.
 * \param fd The file descriptor to use for I/O.
 * \param tv The timer that needs to expire before triggering the callback function.
 * \param job_type A set of flags that indicates the job type.
 * \param func The callback function for a worker thread to call when processing the job.
 * \param data Arbitraty data that will be available to the job callback function.
 * \param job The address of a job pointer that will be filled in once the job is allocated.
 *            \c NULL can be passed if a pointer to the job is not needed.
 * \return \c PR_SUCCESS upon success.
 * \return \c PR_FAILURE upon failure.
 * \warn The thread pool will not allow a job to be added when it has been signaled
 *       to shutdown.
 */
PRStatus ns_add_io_timeout_job(struct ns_thrpool_t *tp,
                               PRFileDesc *fd,
                               struct timeval *tv,
                               ns_job_type_t job_type,
                               ns_job_func_t func,
                               void *data,
                               struct ns_job_t **job);

/**
 * Adds a signal job to the thread pool's work queue
 *
 * \param tp The thread pool whose work queue you want to add a signal job to.
 * \param signum The signal number that you want to trigger the callback function.
 * \param job_type A set of flags that indicates the job type.
 * \param func The callback function for a worker thread to call when processing the job.
 * \param data Arbitraty data that will be available to the job callback function.
 * \param job The address of a job pointer that will be filled in once the job is allocated.
 *            \c NULL can be passed if a pointer to the job is not needed.
 * \return \c PR_SUCCESS upon success.
 * \return \c PR_FAILURE upon failure.
 * \warn The thread pool will not allow a job to be added when it has been signaled
 *       to shutdown.
 */
PRStatus ns_add_signal_job(ns_thrpool_t *tp,
                           PRInt32 signum,
                           ns_job_type_t job_type,
                           ns_job_func_t func,
                           void *data,
                           struct ns_job_t **job);

/**
 * Adds a plain job to the thread pool's or the thread's work queue.  A plain job is a job that
 * is executed immediately that is not contingent on an event or signal.  This is typically
 * used when the server reads a request, and the request cannot be processed in the
 * I/O thread.  The read callback will create a job to perform the task in a separate thread.
 * If there are going to be multiple tasks that must be performed sequentially, the
 * caller can acquire a thread for it's own use and add jobs to only that thread, to
 * guarantee that the jobs will be processed in the order added.
 *
 * \param tp The thread pool whose work queue you want to add a signal job to.
 * \param thr If not NULL, use this thread to process the jobs (for sequential processing)
 * \param job_type A set of flags that indicates the job type (usually just NS_JOB_THREAD)
 * \param func The callback function for a worker thread to call when processing the job.
 * \return \c PR_SUCCESS upon success.
 * \return \c PR_FAILURE upon failure.
 * \warn The thread pool will not allow a job to be added when it has been signaled
 *       to shutdown.
 */
PRStatus ns_add_job(ns_thrpool_t *tp, ns_thread_t *thr, ns_job_type_t job_type, ns_job_func_t func, void *data);

/**
 * Allows the callback to access the private data field in the job.
 *
 * \param job The job to get the data for.
 * \return \c The private data associated with the job.
 */
void *ns_job_get_data(struct ns_job_t *job);

/**
 * Allows the callback to access the thread pool that the job is associated with.
 * Useful for adding jobs from within jobs.
 *
 * \param job The job to get the thread pool for.
 * \return \c The thread pool associated with the job.
 */
ns_thrpool_t *ns_job_get_tp(struct ns_job_t *job);

/**
 * Jobs can be associated with both an I/O event and a timeout event.
 * The callback func may need to know which event triggered the
 * callback.  This function will allow access to the type of event
 * that triggered the callback.  Use NS_JOB_IS_TIMER() to tell if
 * the event is a timer event, like this:
 *   if (NS_JOB_IS_TIMER(ns_job_get_output_type(job))) {
 *      ... handle timeout ...
 *   } else {
 *      ... handle I/O ...
 *   }
 * \param job The job to get the thread pool for.
 * \return \c The ns_job_type_t corresponding to the event that triggered the callback
 */
ns_job_type_t ns_job_get_output_type(struct ns_job_t *job);

/**
 * Creates a new thread pool
 *
 * \param initial_threads The initial number of threads to spawn when creating the pool.
 * \param max_threads The maximum number of threads that can be held in the pool.
 * \param stacksize The stacksize to use for each thread in the pool.
 * \param queue_size The size of the thread pool's work queue.
 * \return A pointer to the newly created thread pool.
 */
struct ns_thrpool_t *ns_thrpool_new(PRInt32 initial_threads,
                                    PRInt32 max_threads,
                                    PRUint32 stacksize,
                                    PRUint32 queue_size);

/**
 * Frees a thread pool from memory
 *
 * This will free a thread pool and it's internal resources from memory.  You should
 * be sure that the thread pool has been shutdown before destroying it by calling
 * ns_thrpool_wait().
 *
 * \param tp The thread pool to destroy.
 * \see ns_thrpool_new()
 * \see ns_thrpool_wait()
 */
void ns_thrpool_destroy(struct ns_thrpool_t *tp);

/**
 * Tells a thread pool to shutdown it's threads
 *
 * \param tp The thread pool to shutdown.
 * \warn This must only be called from a job that runs in a worker thread.  Calling
 *       this function from the event thread can cause a deadlock.
 */
void ns_thrpool_shutdown(struct ns_thrpool_t *tp);

/**
 * Checks if a thread pool is shutting down
 *
 * This can be called by worker threads so they know when the thread pool has
 * been requested to shut down.
 *
 * \return \c 0 if the thread pool is not shutting down.
 * \return \c 1 if the thread pool is shutting down.
 */
PRInt32 ns_thrpool_is_shutdown(struct ns_thrpool_t *tp);

/**
 * Waits for all threads in the pool to exit
 *
 * This call will block the caller until all threads in the thread pool have exited.  A
 * program will typically create the thread pool by calling ns_thrpool_new(), then it
 * will call ns_thrpool_wait() to wait until the thread pool is shutdown (which is likely
 * initiated by a signal handler).  Once this function successfully returns, the thread
 * pool can be safely destroyed by calling ns_thrpool_destroy().
 *
 * \param tp The thread pool to wait for.
 * \return \c PR_SUCCESS on success.
 * \return A NSPR error code on failure.
 * \see ns_thrpool_new()
 * \see ns_thrpool_shutdown()
 * \see ns_thrpool_destroy()
 */
PRStatus ns_thrpool_wait(struct ns_thrpool_t *tp);

/**
 * Get a worker thread from the threadpool for the private use of the caller.
 * This is useful when the caller must be able to process incoming
 * requests in the order received.  Since only one thread operates on the
 * job queue, the jobs are guaranteed to be processed in the order added.
 *
 * \param tp The thread pool from which you want to acquire a free thread
 * \param block If true, wait until there is a free thread - if false, do not wait
 * \return \c a thread object
 * \return \c NULL - if block is true, only return NULL if shutting down - if block
 *                   is false, return NULL if no free threads are available
 */
ns_thread_t *ns_thread_acquire(ns_thrpool_t *tp, PRBool block);

/**
 * Release the worker thread back to the threadpool.
 *
 * \param thr The thread to put back into the free thread list
 */
void ns_thread_release(ns_thread_t *thr);

/*
 * This is a convenience function - use if you need to re-arm the same event -
 * e.g. in the listen cb to re-arm the listen - in a read/write to re-arm if need more data
 */
void ns_job_rearm(struct ns_job_t *job);
void ns_job_modify(struct ns_job_t *job, ns_job_type_t event_type);

#endif /* NS_THRPOOL_H */
