
#include "nspr.h"
#include "plgetopt.h"
#include "plstr.h"
#include "private/pprio.h"
#include <event.h>

struct listen_ctx {
    PRFileDesc *listenfd;
    struct event *ev;
};

#define LEN_BYTES 4

struct conn_ctx {
    PRFileDesc *connfd;
    struct event *ev;
    size_t offset;
    size_t len;
#define CONN_BUFFER_SIZE BUFSIZ
    char buffer[CONN_BUFFER_SIZE];
};

static struct listen_ctx *
listen_ctx_new(PRFileDesc *listenfd, struct event *ev)
{
    struct listen_ctx *listenctx = PR_NEW(struct listen_ctx);
    listenctx->listenfd = listenfd;
    listenctx->ev = ev;
    return listenctx;
}

static struct conn_ctx *
conn_ctx_new(PRFileDesc *connfd, struct event *ev)
{
    struct conn_ctx *connctx = PR_NEW(struct conn_ctx);
    connctx->connfd = connfd;
    connctx->ev = ev;
    connctx->offset = 0;
    connctx->len = 0;
    return connctx;
}

static void
conn_ctx_done(struct conn_ctx *connctx)
{
    if (connctx->ev) {
	event_del(connctx->ev);
	PR_DELETE(connctx->ev);
    }
    if (connctx->connfd) {
	PR_Close(connctx->connfd);
	connctx->connfd = NULL;
    }
    connctx->offset = 0;
}

static void
conn_ctx_free(struct conn_ctx **connctx)
{
    conn_ctx_done(*connctx);
    PR_DELETE(*connctx);
}

static void
conn_write(int fd, short event, void *arg)
{
    struct conn_ctx *connctx = (struct conn_ctx *)arg;
    PRInt32 len, nbytes;
    short evtype = 0;
    
    PR_ASSERT(arg);
    PR_ASSERT(fd == PR_FileDesc2NativeHandle(connctx->connfd));
    nbytes = connctx->len-connctx->offset;
    len = PR_Write(connctx->connfd, connctx->buffer+connctx->offset, nbytes);
    if (len < 0) {
	PRErrorCode prerr = PR_GetError();
	(void)prerr; /* TODO */
	/* error */
	conn_ctx_free(&connctx);
    } else if (len == 0) {
	/* closed */
	conn_ctx_free(&connctx);
    } else {
	connctx->offset += len;
	if (connctx->offset < connctx->len) {
	    /* reschedule write */
	    evtype = EV_WRITE;
	} else if (connctx->offset == connctx->len) {
	    /* done */
	    /* delete event */
	    conn_ctx_free(&connctx);
	} else {
	    /* buffer overflow error */
	}
    }
    if (evtype) {
	event_add(connctx->ev, NULL);
    }
}

static void
conn_read(int fd, short event, void *arg)
{
    struct conn_ctx *connctx = (struct conn_ctx *)arg;
    PRInt32 len, nbytes;
    short evtype = 0;
    
    PR_ASSERT(arg);
    PR_ASSERT(fd == PR_FileDesc2NativeHandle(connctx->connfd));
    if (connctx->len) {
	nbytes = connctx->len-connctx->offset;
    } else {
	nbytes = CONN_BUFFER_SIZE-connctx->offset;
    }
    len = PR_Read(connctx->connfd, connctx->buffer+connctx->offset, nbytes);
    if (len < 0) {
	PRErrorCode prerr = PR_GetError();
        (void)prerr; /* TODO */
	/* error */
	conn_ctx_free(&connctx);
    } else if (len == 0) {
	/* closed */
	conn_ctx_free(&connctx);
    } else {
	connctx->offset += len;
	if (!connctx->len && (connctx->offset >= LEN_BYTES)) {
	    PRUint32 *val = (PRUint32 *)connctx->buffer;
	    connctx->len = (PRInt32)PR_ntohl(*val);
	    nbytes = connctx->len-connctx->offset;
	}
	if (len < nbytes) {
	    /* reschedule read */
	    evtype = EV_READ;
	} else {
	    /* write now */
	    evtype = EV_WRITE;
	    connctx->offset = 0;
	    event_set(connctx->ev, PR_FileDesc2NativeHandle(connctx->connfd), evtype, conn_write, connctx);
	}
    }
    if (evtype) {
	event_add(connctx->ev, NULL);
    }
}

static void
listen_accept(int fd, short event, void *arg)
{
    PRFileDesc *connfd = NULL;
    struct listen_ctx *listenctx = (struct listen_ctx *)arg;
    struct event *ev;
    struct conn_ctx *connctx;
    PRSocketOptionData prsod = {PR_SockOpt_Nonblocking, {PR_TRUE}};

    PR_ASSERT(arg);
    PR_ASSERT(fd == PR_FileDesc2NativeHandle(listenctx->listenfd));

    /* accept and add read event for new connection */
    connfd = PR_Accept(listenctx->listenfd, NULL, PR_INTERVAL_NO_TIMEOUT);
    if (connfd) {
	PR_SetSocketOption(connfd, &prsod);
	ev = PR_NEW(struct event);
	connctx = conn_ctx_new(connfd, ev);
	event_set(ev, PR_FileDesc2NativeHandle(connfd), EV_READ, conn_read, connctx);
	event_add(ev, NULL);
    }
    /* re-arm listen_accept */
    event_add(listenctx->ev, NULL);

    return;
}

int
main (int argc, char **argv)
{
    char *progname = argv[0];
    argv++;

    (void)progname; /* TODO */
    /* Initalize the event library */
    event_init();

    /* arguments are listen ports and/or unix domain listen file names */
    for (; argv && *argv; ++argv) {
	struct listen_ctx *listenctx;
	struct event *ev;
	PRFileDesc *listenfd = NULL;
	PRNetAddr netaddr;
	int portnum = atoi(*argv);
	PRSocketOptionData prsod = {PR_SockOpt_Nonblocking, {PR_TRUE}};

	if (portnum == 0) { /* unix socket */
	    netaddr.local.family = PR_AF_LOCAL;
	    PL_strncpyz(netaddr.local.path, *argv, sizeof(netaddr.local.path));
	} else {
	    netaddr.inet.family = PR_AF_INET;
	    netaddr.inet.port = PR_htons(portnum);
	    netaddr.inet.ip = PR_htonl(PR_INADDR_ANY);
	}
	listenfd = PR_OpenTCPSocket(PR_NetAddrFamily(&netaddr));
	PR_SetSocketOption(listenfd, &prsod);
	prsod.option = PR_SockOpt_Reuseaddr;
	PR_SetSocketOption(listenfd, &prsod);
	PR_Bind(listenfd, &netaddr);
	PR_Listen(listenfd, 32);

	/* create the event for accepting new connections from this socket */
	ev = PR_NEW(struct event);
	listenctx = listen_ctx_new(listenfd, ev);
	
	event_set(ev, PR_FileDesc2NativeHandle(listenfd), EV_READ, listen_accept, listenctx);
	event_add(ev, NULL);
    }

    event_dispatch();

    return (0);
}

